"File PText.st
======================================================================
|
| Copyright (C) 1995 Free Software Foundation, Inc.
| Written by Brad Diller.
|
| This file is part of GNU Smalltalk.
|
| GNU Smalltalk is free software; you can redistribute it and/or modify it
| under the terms of the GNU General Public License as published by the Free
| Software Foundation; either version 1, or (at your option) any later version.
| 
| GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
| ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
| FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
| details.
| 
| You should have received a copy of the GNU General Public License along with
| GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
| Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  
|
 ======================================================================
 

|     Change Log
| ============================================================================
| Author        Date              Change
| bdiller       18 Sep 93         Created.
| bdiller       12 Jul 95         Modified extensively to support
|                                 revamped window management system.
"
Primitive subclass:  #PText
          instanceVariableNames: ' textMsg selection'
          classVariableNames: 'buffer'
          poolDictionaries: ''
          category: 'Graphics Windows'
!

!PText class methodsFor: 'instance creation'!

new: parent 
| view |
    view := self new.
    view parentView: parent.
    view blox: (BText new: parent blox).
    view blox contents: ''.
    view blox setCallback: (view blox guiObject) receiver: view message:    'setDirtyFlag'.
   ^view.
!

newReadOnly: parent
| view |
    view := self new.
    view parentView: parent.
    view blox: (BText newReadOnly: parent blox).
    view blox contents: ''.
    ^view.
!

initBuffer
	buffer := nil
!!

!PText methodsFor: 'initializing'!

stateChange: stateChangeKey
"Install message handler to redraw text pane in response to an stateChangeKey 
message.  If there is text which is initially selected, select the text.  This
feature is utilized by some types of message set browsers"
    self stateChange: stateChangeKey updateWith: 
	[ self display.
	  selection notNil
	      ifTrue: [self findString: selection]].
!

selection: aString
    selection := aString
!

setDirtyFlag
"Set modification state of text view"
    dirty := true
!

textMsg: textSelector
"The textSelector is supplied by the view's data object.  When invoked
from computeText, the text to be displayed is returned"
    textMsg := textSelector
!!

!PText methodsFor: 'displaying'!

displayError: errorString at: lineNo
"Display error string at end of line indicated by lineNo"
    ((self blox gotoLine: (self blox guiObject) line: lineNo end: true) = 0)
	ifFalse: [self blox insertSelectedText: errorString]
	ifTrue: [self beep].
!

displayError: errorString
"Insert error string at cursor and select it"
    self blox insertTextSelection: errorString
!

display
"Update text view.  Dirty flag is reset"
    blox contents: (data perform: textMsg).
    dirty := false.
!

findString: aString
"Select aString in the text view.  If not found, beep"
    ((blox  searchString: (blox guiObject) string: aString) = 0)
	ifTrue: [self beep].
!

confirm: aString
    "Used by canUpdate when the text has been modified.  If the user wishes to 
     discard the editing changes by pressing yes, the dirty flag is reset"
    ModalDialog new message: aString in: self;
	addButton: 'Yes' message: [dirty := false];
	addButton: 'No' message: [];
	display: self.
    ^dirty not
!

canUpdate
"If text has been modified, display a prompter.  If the No button is
selected, return true"
    | cancel |
    data isNil ifTrue: [^true].
    dirty ifFalse: [^true].
    cancel := self confirm: 'The text has been altered.' , 
	Character nl asSymbol , 'Do you wish to discard those changes?'.
    ^cancel.
!!


!PText methodsFor: 'blue button menu items'!

gstCut
    self gstCopy.
    self gstClear.
!

gstCopy
    buffer := blox getSelection.
!


gstPaste
  (buffer isNil) ifTrue: [^self beep].
    blox getSelection = nil 
	ifTrue: [blox insertText: '']
	ifFalse: [blox replaceSelection: buffer].
!

gstClear
    blox replaceSelection: ''.
!

line
"Prompt user to enter a line number.  If a valid number, attempt
to scroll to entered line number"
| prompter alert |
    prompter := Prompter message: 'Goto line...' in: self.
    prompter response ~= '' 
	ifTrue: [
		 prompter response do: 
		     [ :ch |
			    (ch isDigit) 
				"Need to anchor this alert to a temporary variable to prevent the object from
				 being GC'd "
				ifFalse: [^self blox beep]].
		 ((blox gotoLine: (blox guiObject) line: (prompter response 
							      asInteger) end: false) ~= 0 )   ifFalse: [self beep]].  
!

find
| prompter |
    prompter := Prompter message: 'Search...' in: self.
    prompter response ~= '' 
	ifTrue: [blox searchString: (blox guiObject) string: prompter response].  
!

eval
| text |
    text := blox getSelection.
    (text isNil or: [text size = 0]) ifTrue: [^self beep].
    Behavior  evaluate:  text.
!

evalAndPrintResult
"Display and select result of evaluation of selected expression to right of
selection"
| pos text aStream |
    text := blox getSelection.
    (text isNil or: [text size = 0]) ifTrue: [^self beep].
    aStream := WriteStream on: (String new: 0).
    (Behavior  evaluate:  text ifError: [:fname :lineNo :errorString | 
								       errorString asSymbol ]) printOn: aStream.
    blox insertTextSelection: aStream contents.
!

evalAndInspectResult
"Open an inspector on the result of the evaluation of the selected Smalltalk expression"
| obj text|
    text := blox getSelection.
    (text isNil or: [text size = 0]) ifTrue: [^self beep].
    obj := (Behavior  evaluate:  text ifError: [:fname :lineNo :errorString | 
									      self displayError: errorString. ^nil ]).
    obj inspect.
!


compileIt 
"Activated when the user selects 'accept' from the popup menu.  Send a 
stateChangeMsg to the data object"
    | text |
		text := blox contents.
    (text isNil or: [text size = 0]) ifTrue: [^self beep].
    self canChangeState ifTrue:
	[stateChangeMsg numArgs = 1
	     "One parameter selector"
	     ifTrue:
		 [dirty := (data perform: stateChangeMsg with: blox contents) isNil]
	     "Two parameter selector"
	     ifFalse:
		 [dirty := (data perform: stateChangeMsg with: (blox contents) with: self) isNil]]
!

revert
    "Revert text changes and replace current text with original text" 
    self display.
!!

